/* 
 * Please read the COPYRIGHT notice included at the top level
 */

/*
 * instr.c,v 1.1.1.1 1996/07/18 22:39:55 arsharma Exp
 * Purpose: To do some per instruction processing like checking if the
 * instruction needs to be augmented etc
 */

#include "doctor_t.h"
#include "lex.h"
#include "memref.h"
#include "instr.h"
#include "write.h"
#include "augment.h"
#include "mnemonics_def.h"
#include "registers.h"

#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>

extern int instr_num;
extern unsigned static_execution_time;

/* another circular buffer of instructions */
extern instr_t **basic_block;

extern bool_t augment_on;

int start_block = 0;

int instr_count;

/* the -i command line flag */
extern bool_t instr_count_only;

/*
 * augmentation       buffer. This is flushed, at the end of every basic
 * block
 */
char *aug_buffer;

/* Allocates and initializes an instruction structure */
instr_t *
new_instruction ()
{
    instr_t *new_instr;

    new_instr = basic_block[instr_num] = (instr_t *) malloc (sizeof
							     (instr_t));
    /* clean it up */
    memset ((void *) new_instr, 0, sizeof (instr_t));

    instr_num = SUCC_INSTR (instr_num);

    return new_instr;
}

/* look in write.c for more info */
extern char *aug_buffer;
extern int aug_end;

/* flush the last seen  basic block */
int
flush_basic_block ()
{
    int i;
    enum enum_ref_t ref;
    instr_t *instr;

    /*
     * instrnum       is the beginning of the new basic block. Flush
     * everything  before that
     */
    for (i = start_block; i != instr_num; i = SUCC_INSTR (i)) {
	char *aug;

	instr = basic_block[i];

	/* increment the instruction count */
        instr_count++;

	/* flush everything before the current instruction */
	flush (instr->start_instr);

	/* do some misc per instruction stuff */
	process_instruction (instr);

	/*
	 * check to see       if there was some augmentation done before the
	 * first instruction. If so, that needs to be printed out now.
	 */
	if ((i == start_block) && instr->augment)
	    fprintf (stdout, "%s", aug_buffer);

	/*
	 * all augmentation for       this instruction will be put in
	 * starting here
	 */
	instr->aug_start = aug_buffer + aug_end;

	/* check the instruction for memory references */
	check_mem_ref (instr);

	if (not(instr_count_only)) {
            /*  augment if necessary */
            if ((MEM_REF (instr)) && (instr->augment))
                augment_instruction (instr);
        }

	if (i == PRED_INSTR (instr_num)) {
            if (instr_count_only)     		
                augment_update_instr_count();
	    else
	        /* update the clock */ 
	        augment_safe_update_clock ();
        }

	/* write out the augmentation */
	fprintf (stdout, "%s", instr->aug_start);

	/* flush the current instruction */
	flush (end_line (instr->start_instr));

	if (instr->post_aug_start)
	    fprintf (stdout, "%s", instr->post_aug_start);
	free (instr);
    }

    /*
     * flush the augmentation buffer if       it was non-empty. Otherwise,
     * retain the augmentation in the buffer, till a non-empty block
     * is encountered
     */
    if (i != start_block)
	flush_aug ();

    /* make a fresh start */
    start_block = instr_num;
}

/* clean up by flushing the last remaining basic block */
int
cleanup ()
{
    /* some Assembly programs may be without instructions ! */
    if (instr_num)
	flush_basic_block ();

    /* flush whatever is remaining in the buffer */
    flush (save_pos ());
}

/*
 * Does       miscellaneous per instruction processing.
 *   - increments the clock
 *   - sets the register mask
 */
void
process_instruction (instr_t *instr)
{
    /* can't handle repeat prefixes yet */
    if (instr->rep)
	instr->augment = FALSE;

    if (instr->augment)
	static_execution_time += instr->mnemonic->clock_cycles;

    instr->double_data = (instr->mnemonic->operand_size == 8);

    /* set the implicit register usage masks */
    switch (instr->mnemonic->representation) {
	case CMPSB:
	case CMPSL:
	case CMPSW:
	    instr->reg_mask |= (DS_MASK | ESI_MASK | ES_MASK | EDI_MASK);
	    break;
	case ENTER:
	case LEAVE:
	case CALL:
	case LCALL:
	case RET:
	case LRET:
	case INT:
	case IRET:
	case IRETF:
	case LIRET:
	case POP:
	case POPF:
	case POPFL:
	case POPL:
	case PUSH:
	case PUSHF:
	case PUSHAW:
	case PUSHFL:
	case PUSHFW:
	case PUSHL:
	case PUSHW:
	    instr->reg_mask |= (SS_MASK | ESP_MASK);
	    break;
	case POPA:
	case POPAL:
	case PUSHA:
	case PUSHAL:
	    /* They touch all registers */
	    instr->reg_mask = 0xffffffff;
	    break;
	case INS:
	case INSB:
	case INSW:
	case INSL:
	case STOS:
	case STOSB:
	case STOSW:
	case STOSL:
	    instr->reg_mask |= (ES_MASK | EDI_MASK);
	    break;
	case LODS:
	case LODSB:
	case LODSW:
	case LODSL:
	case OUTS:
	case OUTSB:
	case OUTSW:
	case OUTSL:
	    instr->reg_mask |= (DS_MASK | ESI_MASK);
	    break;
	default:
	    break;
    }
}
