/* 
 * Please read the COPYRIGHT notice included at the top level
 */
/*
 * Definitions of icodes, threads, queues, and locks. Macros for
 * address mapping, and register dereferencing.
 *
 * Copyright (C) 1993 by Jack E. Veenstra (veenstra@cs.rochester.edu)
 * 
 * This file is part of MINT, a MIPS code interpreter and event generator
 * for parallel programs.
 * 
 * MINT is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 1, or (at your option)
 * any later version.
 * 
 * MINT is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with MINT; see the file COPYING.  If not, write to
 * the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#define _BSD_SIGNALS
/* #include <signal.h>   ($Id)$ */
#include <sys/types.h>
#include <unistd.h>

#ifdef MIPSEL
#define DECSTATION 1
#define host_mips 1
#define CONST
#endif

#ifdef MIPSEB
#define host_mips 1
#endif

#if defined(DECSTATION) || (defined(sparc) && !defined(__svr4__))
typedef unsigned long ulong;
#endif

#ifndef CONST
#define CONST const
#endif

typedef struct thread *thread_ptr;

#include "event.h"
#include "task.h"
#include "hash.h"
#include "export.h"

/* template for event functions */
#define OP(NAME)    void NAME(task_ptr ptask, thread_ptr pthread)

/* The possible run states of a simulated process */
typedef enum {
    R_RUN = 0,			/* start with 0 for efficiency */
    R_SLEEP,			/* not runnable because it caused an event */
    R_STALL,			/* stalled waiting for a pending load
                                 * into a register */
    R_BLOCK,			/* blocked on a lock or semaphore */
    R_WAIT,			/* waiting for children to finish */
    R_ZOMBIE,			/* almost dead; next execution will
                                 * change it to R_DONE */
    R_DONE,			/* waiting for parent to wait on me */
    R_FREE,			/* available for reuse */
    R_MAX			/* must be last; not currently used */
} state_t;

#define WAKEUP(T) ((T)->runstate = R_RUN)
#define SLEEP(T) ((T)->runstate = R_SLEEP)
#define STALL(T) ((T)->runstate = R_STALL)

/*
 * Magic values to check validity of locks and semaphores passed in
 * as arguments.
 */
#define MAGIC_LOCK	0xdeadca1f
#define MAGIC_SEMA	0xc001babe
#define MAGIC_BARRIER	0xbadf00d

/*
 * A lock is an integer value plus a wait queue. The value is 0 if the lock
 * is free, otherwise it is 1. Locks block a process unless spinning is
 * explicitly requested on the command line.
 */
typedef struct uslock_t {
    int val;
    qnode_t wait_q;
    int pid;			/* last process to set this lock */
    int magic;
    /* next and prev are used to link allocated locks on a global list */
    struct uslock_t *next;
    struct uslock_t *prev;
} uslock_t, *uslock_ptr;

/*
 * A semaphore is a integer value plus a wait queue. The value is a counter
 * that is decremented on every P() operation and incremented on every V()
 * operation. If a P() operation decrements the counter to a negative value,
 * the P() operation blocks. If a V() operation finds the counter negative,
 * then one waiting process is unblocked.
 *
 * Barriers are also implemented with semaphores.
 */
typedef struct sema_t {
    int val;
    qnode_t wait_q;
    int magic;
    /* next and prev are used to link allocated semaphores on a global list */
    struct sema_t *next;
    struct sema_t *prev;
} sema_t, *sema_ptr;

/*
 * If you change the thread structure, you must change the constants
 * in ctxsw.h
 */
typedef struct thread {
    thread_ptr next;		/* forward link for queues, MUST BE FIRST */
    thread_ptr prev;		/* backward link for queues, MUST BE SECOND */
    long reg[32];		/* an extra register for writes to r0 */
    long lo;
    long hi;
    float fp[32];		/* floating point (and double) registers */
    long fcr0;			/* floating point control register 0 */
    long fcr31;			/* floating point control register 31 */
    event_ptr pevent;		/* pointer to an event structure */
    state_t runstate;		/* status of thread (runnable,
                                 * sleeping, etc.) */
    int pid;			/* process id */
    long paddr;			/* physical address computed by an
                                 * event function */
    mint_time_t time;		/* simulated time for this thread */
    mint_time_t cpu_time;	/* accumulated cpu time for this thread */
    double child_cpu;		/* accumulated cpu time for finished
                                 * children */
    int is_zombie;		/* =1 if thread is almost dead */
    long addr_space;		/* start of allocated private data in
                                 * MINT space */
    thread_ptr parent;		/* pointer to parent */
    thread_ptr youngest;	/* pointer to youngest child */
    thread_ptr sibling;		/* pointer to next older sibling */
    int (*ufunc) ();		/* user-defined or event-specific
                                 * function to call */
    int index;			/* index to fevent (fake events) */
    int started;		/* used for INTEL. if 0 ctxsw_first is used */
    int num_wait_for_end;	/* used with WAIT_FOR_END.  can be negative */
#ifdef ULONG_TIME
    char pad[40];		/* pad thread struct so that size is
                                 * power of 2 */
#else
    char pad[32];		/* pad thread struct so that size is
                                 * power of 2 */
#endif
    double tmp1, tmp2;		/* temporaries used in data movement */
} thread_t;

/*
 * MAP() translates addresses from virtual (the object file's address space)
 * to physical (the real address space used in MINT)
 */
#define MAP(ADDR) ((long) (ADDR))

/* PMAP() is used when the address is known to be private */
#define PMAP(ADDR) ((long) (ADDR))
#define PUNMAP(ADDR) ((long) (ADDR))

/* returns true if a virtual address is shared */
#define IS_SHARED(ADDR) ((long) (ADDR) >= Shmem_start && (long) (ADDR) < Shmem_end)

/* SMAP() is used when the address is known to be shared */
#define SMAP(ADDR) ((long) (ADDR))

/* SUNMAP() is used when the address is known to be shared */
#define SUNMAP(ADDR) ((long) (ADDR))

/* converts a physical address back to the objects virtual address */
#define UNMAP(ADDR) ((long) (ADDR) >= Shmem_start ? SUNMAP(ADDR) : PUNMAP(ADDR))

#define FORK_CYCLES        1
#define UNLOCK_CYCLES      1
#define BARRIER_CYCLES     1
#define VSEMA_CYCLES       1
#define TERMINATE_CYCLES   1

/* typedef void (*PFV)();  declared in common.h */

OP (event_error);
OP (event_read);
OP (event_write);
OP (event_prefork);
OP (event_fork);
OP (event_new_lock);
OP (event_init_lock);
OP (event_lock_attempt);
OP (event_lock_block);
OP (event_lock_acquire);
OP (event_lock_release);
OP (event_lock_unblock);
OP (event_free_lock);
OP (event_new_barrier);
OP (event_init_barrier);
OP (event_barrier_attempt);
OP (event_barrier_block);
OP (event_barrier_acquire);
OP (event_free_barrier);
OP (event_new_sema);
OP (event_init_sema);
OP (event_psema_attempt);
OP (event_psema_block);
OP (event_psema_acquire);
OP (event_vsema);
OP (event_vsema_unblock);
OP (event_free_sema);
OP (event_shmalloc);
OP (event_shfree);
OP (event_malloc);
OP (event_free);
OP (event_terminate_task);
OP (event_terminate_thr);
OP (event_done_thr);
OP (event_user);

int main_envelope ();
int child_envelope ();
int next_event (task_ptr ptask);
