/* 
 * Please read the COPYRIGHT notice included at the top level
 */
/*
 * Routines for the main execution loop, thread queues, and all the event
 * generating functions.
 *
 * Copyright (C) 1993 by Jack E. Veenstra (veenstra@cs.rochester.edu)
 * 
 * This file is part of MINT, a MIPS code interpreter and event generator
 * for parallel programs.
 * 
 * MINT is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 1, or (at your option)
 * any later version.
 * 
 * MINT is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with MINT; see the file COPYING.  If not, write to
 * the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

#include <stdio.h>
#include "icode.h"
#include "globals.h"

/* function prototypes */
void init_q (qnode_ptr qhead);
int onqueue (qnode_ptr qhead, qnode_ptr pthread);
void sleep_thr (thread_ptr pthread);
void wakeup_thr (thread_ptr pthread);
OP (event_wait_for_end);
OP (event_check_tasks);
OP (event_check_children);
OP (event_check_parent);
OP (event_done_thr);
OP (event_synch_time);
void init_all_queues ();
void init_main_thread ();

void task_insert (task_ptr ptask);

/*
 * Table for mapping of event type as defined in event.h to event
 * functions to be executed in the next_event routine.  To access this
 * table, the event type must be shifted right by 13 bits (use the
 * TABLE_OFFSET inline macro).
 * Note:  The comments on the right correspond to the numbers assigned
 * to the event types in file event.h.  If the definitions for the
 * types change, the table must also be updated.
 */
PFV Event_tab[] =
{
    event_error,		/* 0 */

    event_read,			/* 1 */
    event_write,

    event_prefork,		/* 3 */
    event_fork,

    event_new_lock,		/* 5 */
    event_init_lock,
    event_lock_attempt,
    event_lock_block,
    event_lock_acquire,
    event_lock_release,
    event_lock_unblock,
    event_free_lock,

    event_new_barrier,		/* 13 */
    event_init_barrier,
    event_barrier_attempt,
    event_barrier_block,
    event_barrier_acquire,
    event_free_barrier,

    event_new_sema,		/* 19 */
    event_init_sema,
    event_psema_attempt,
    event_psema_block,
    event_psema_acquire,
    event_vsema,
    event_vsema_unblock,
    event_free_sema,

    event_shmalloc,		/* 27 */
    event_shfree,
    event_malloc,
    event_free,

    event_wait_for_end,		/* 31 */

    event_check_tasks,		/* 32 */
    event_check_children,
    event_check_parent,
    event_done_thr,

    event_user,			/* 36 */

    event_synch_time,		/* 37 */
};

int
main_envelope ()
{
    appl_main (appl_argc, appl_argv);

    pcur_thread->pevent->type = E_SYNCH_TIME;
    (*ctxsw_fn) (pcur_thread, psim_thread, &pcur_thread);

    pcur_thread->pevent->type = E_CHECK_TASKS;
    (*ctxsw_fn) (pcur_thread, psim_thread, &pcur_thread);

    pcur_thread->pevent->type = E_SYNCH_TIME;
    (*ctxsw_fn) (pcur_thread, psim_thread, &pcur_thread);

    pcur_thread->pevent->type = E_CHECK_CHILDREN;
    (*ctxsw_fn) (pcur_thread, psim_thread, &pcur_thread);

    pcur_thread->pevent->type = E_SYNCH_TIME;
    (*ctxsw_fn) (pcur_thread, psim_thread, &pcur_thread);

    pcur_thread->pevent->type = E_CHECK_PARENT;
    (*ctxsw_fn) (pcur_thread, psim_thread, &pcur_thread);

    pcur_thread->pevent->type = E_SYNCH_TIME;
    (*ctxsw_fn) (pcur_thread, psim_thread, &pcur_thread);

    pcur_thread->pevent->type = E_DONE_THR;
    (*ctxsw_fn) (pcur_thread, psim_thread, &pcur_thread);
}

int
child_envelope ()
{
    (*pcur_thread->ufunc) ();

    pcur_thread->pevent->type = E_SYNCH_TIME;
    (*ctxsw_fn) (pcur_thread, psim_thread, &pcur_thread);

    pcur_thread->pevent->type = E_CHECK_TASKS;
    (*ctxsw_fn) (pcur_thread, psim_thread, &pcur_thread);

    pcur_thread->pevent->type = E_SYNCH_TIME;
    (*ctxsw_fn) (pcur_thread, psim_thread, &pcur_thread);

    pcur_thread->pevent->type = E_CHECK_CHILDREN;
    (*ctxsw_fn) (pcur_thread, psim_thread, &pcur_thread);

    pcur_thread->pevent->type = E_SYNCH_TIME;
    (*ctxsw_fn) (pcur_thread, psim_thread, &pcur_thread);

    pcur_thread->pevent->type = E_CHECK_PARENT;
    (*ctxsw_fn) (pcur_thread, psim_thread, &pcur_thread);

    pcur_thread->pevent->type = E_SYNCH_TIME;
    (*ctxsw_fn) (pcur_thread, psim_thread, &pcur_thread);

    pcur_thread->pevent->type = E_DONE_THR;
    (*ctxsw_fn) (pcur_thread, psim_thread, &pcur_thread);
}

void
init_all_queues ()
{
    int pid;

    /* init the queue head nodes */
    INLINE_INIT_Q (&Run_q);
    INLINE_INIT_Q (&Done_q);
    INLINE_INIT_Q (&Free_q);
    INLINE_INIT_Q (&Sleep_q);
    INLINE_INIT_Q (&Sema_list);
    INLINE_INIT_Q (&Lock_list);

    /*
     * Initialize the simulator thread.  We don't use the Threads
     * array because it is for application threads.
     */
    psim_thread = (thread_ptr) calloc (1, sizeof (thread_t));
    if (psim_thread == NULL)
	fatal ("init_all_queues: cannot allocate 0x%x bytes for sim thread.\n",
	       sizeof (thread_t));
    psim_thread->pid = -1;
    psim_thread->addr_space = 0;

    /*
     * Allocate and initialize the array of threads.
     */
    Threads = (thread_ptr) calloc (Max_nprocs, sizeof (thread_t));
    if (Threads == NULL)
	fatal ("init_all_queues: cannot allocate 0x%x bytes for threads.\n",
	       Max_nprocs * sizeof (thread_t));
    for (pid = 0; pid < Max_nprocs; pid++) {
	Threads[pid].pid = pid;
	Threads[pid].addr_space = 0;
	INLINE_ENQUEUE (&Free_q, &Threads[pid]);
    }
}

void
init_event (thread_ptr pthread, event_ptr pevent)
{
    pthread->pevent = pevent;
    pevent->cpu_time = &pthread->cpu_time;
    pevent->pid = pthread->pid;
    pevent->next = NULL;
}

void
init_thread (thread_ptr pthread, event_ptr pevent)
{
    int i;
    sema_ptr psema;

    pthread->reg[0] = 0;
    pthread->cpu_time = 0;
    pthread->child_cpu = 0;
    pthread->is_zombie = 0;
    pthread->ufunc = NULL;
    pthread->runstate = R_RUN;
    pthread->youngest = NULL;
    pthread->started = 0;
    pthread->num_wait_for_end = 0;

    /* put the thread on the Run queue */
    INLINE_ENQUEUE (&Run_q, pthread);

    /* initialize the event structure associated with this thread */
    init_event (pthread, pevent);
}

void
init_main_thread ()
{
    int i;
    thread_ptr pthread;
    event_ptr pevent;

    Maxpid = 0;
    Nprocs = 1;
    pthread = &Threads[0];
    INLINE_REMOVE (pthread);	/* remove thread from Free_q */
    NEW_ITEM (Event_free, sizeof (event_t), pevent, "init_main_thread");
    init_thread (pthread, pevent);
    pthread->time = 0;
    pthread->parent = NULL;
    pthread->sibling = NULL;
    pthread->fcr31 = 0;
    /*
     * Set the return address to main_envelope so that after a
     * context switch, main_envelop will execute.  It will register
     * the thread before and clean up the thread after the function
     * call pthread->ufunc set below. 
     */
    pthread->reg[31] = (int) main_envelope;
}

/*
 * This event is used to synchronize time before events that require
 * cycle accuracy (e.g. wait_for_end, check_parents, locks, barriers,
 * etc.).  These events "execute" during the event_handler but not in
 * the associated tasks as in most other events (e.g. read and write)
 * that execute at the correct time.  Execution of event handlers do
 * not go through the time wheel and can violate threads' timing
 * dependency.  Thus, without the synch_time event, the former event
 * might be "executed" out of the correct order.  
 */
OP(event_synch_time)
{
    ptask->ufunc = task_synch_time;
    SLEEP(pthread);
}

OP (event_wait_for_end)
{
    event_ptr pevent;

    pevent = pthread->pevent;
    pthread->num_wait_for_end += pevent->arg1;
    if (pthread->num_wait_for_end > 0) {
	/* the WAIT_FOR_END condition is not satisfied */
	/* block myself */
	ptask->ufunc = task_block;
	sleep_thr (pthread);
	pthread->runstate = R_WAIT;
    }
}

OP (event_check_tasks)
{
#ifdef DEBUG_VERBOSE
    debug_print_event (ptask, pthread, pthread->pevent,
                       "event_check_tasks");
#endif /* DEBUG_VERBOSE */
    /* Function to be executed by sched_loop. */
    ptask->ufunc = task_terminate;
    SLEEP (pthread);
}

/*
 * This event checks if all the children
 * have finished, and collects their accumulated time values. If not
 * all children have finished, this thread goes to sleep.  When all
 * the children have finished, this thread will execute event_done to
 * wake up its parent if blocked.
 */
OP (event_check_children)
{
    event_ptr pevent;

#ifdef DEBUG_VERBOSE
    debug_print_event (ptask, pthread, pthread->pevent,
                       "event_check_children");
#endif /* DEBUG_VERBOSE */
    /* check if all the children are done, and free those that are */
    if (wait_for_children (pthread) == 0) {
	/* not all the children have finished */
	/* block myself */
	pevent = pthread->pevent;
	ptask->ufunc = task_block;
	sleep_thr (pthread);
	pthread->runstate = R_WAIT;
    }
}

OP (event_check_parent)
{
    thread_ptr parent;
    mint_time_t duration;
    event_ptr pevent;

    /* if this is the main thread, then all processes have finished */
    if (pthread == &Threads[0]) {
	/*
	 *
	 * Schedule task_yield so that it can return to the thread
	 * envelop and generate event_done_thr.
	 */
	ptask->ufunc = task_yield;
	SLEEP (pthread);
    }
    else {
	/* this is not the main thread */

	/* if parent thread is waiting for us, then wake it up */
	if (wakeup_parent (pthread)) {
	    parent = pthread->parent;

	    /* advance the parent forward in time */
	    duration = pthread->time + TERMINATE_CYCLES - parent->time;
	    if (duration < 0)
		fatal ("event_check_parent: negative wait duration"
                       " (%.1f) at time %.0f\n",
		       duration, pthread->time + TERMINATE_CYCLES);
	    parent->time += duration;
	    pthread->time += TERMINATE_CYCLES;
	    wakeup_thr (parent);

	    /* schedule unblock task */

	    /*
	     *
	     * pthread is a pointer to the unblocking thread.  parent is a
	     * pointer to the unblocked thread.  The dureation and utype
	     * fields of the event structure of the unblocked thread is
	     * set by the caller. 
	     */
	    pevent = pthread->pevent;

	    /*
	     * Set the pevent field in the event structure for the unblocked
	     * thread to point to the unblocking thread's event structure so
	     * that sim_block() can figure out who did the unblocking.
	     */
	    parent->pevent->pevent = pevent;

	    /*
	     * Set the child pevent->pevent to the parent's pevent so that
	     * the child knows whom to unblock.
	     */
	    pevent->pevent = parent->pevent;
	    ptask->ufunc = task_unblock;
	    SLEEP (pthread);
	    parent->pevent->duration = duration;
	}
	pthread->is_zombie = 1;
    }
}

/*
 * This is the last function called by a thread. It moves the thread
 * to the free queue if the parent has already waited on it.
 */
OP (event_done_thr)
{
    event_ptr pevent;

#ifdef DEBUG_VERBOSE
    debug_print_event (ptask, pthread, pthread->pevent, "event_done_thr");
#endif /* DEBUG_VERBOSE */

    pevent = pthread->pevent;

    /*
     * Generate an event so that the task scheduler will decrement Nprocs.
     */
    ptask->ufunc = task_done;
    SLEEP (pthread);

    /*
     * If parent has already waited on this thread, then we can move
     * this thread to the free queue. Otherwise, move this thread to
     * the done queue and the parent will move this thread to the free
     * queue when it waits on this thread.
     */
    if (pthread->parent == NULL) {
	INLINE_REMOVE (pthread);
	if (Recycle_threads)
	    INLINE_INSERT_AFTER (&Free_q, pthread);
	pthread->runstate = R_FREE;
    }
    else {
	INLINE_REMOVE (pthread);
	INLINE_ENQUEUE (&Done_q, pthread);
	pthread->runstate = R_DONE;
    }
}

/*
 * This routine is called when a process terminates in order to wait
 * for all its children to complete. This guarantees that the elapsed
 * time of the parent is at least as great as all of its children.
 * Return values:
 *    -1  if all children have finished
 *     0  if some child has not finished
 */
int
wait_for_children (thread_ptr pthread)
{
    int childpid;

    while ((childpid = wait_for_child (pthread)) > 0);

    return childpid;
}

/*
 * This routine is called by wait() to wait for one child. If there is
 * more than one terminated child, the child that terminated first
 * is the one that is waited on.
 * Return values:
 *   -1   no children exist
 *    0   no child has terminated yet
 *  pid   the child pid of the earliest terminated child process
 */
int
wait_for_child (thread_ptr pthread)
{
    thread_ptr child;
    thread_ptr prev;
    mint_time_t min_time;

    child = pthread->youngest;

    /* if there are no children, return -1 */
    if (child == NULL)
	return -1;

    min_time = MAXTIME;
    /* search the list of children for one that has finished */
    for (; child; child = child->sibling) {
	if (child->runstate == R_DONE || child->is_zombie) {
	    /* find the earliest terminated child */
	    if (min_time > child->time)
		min_time = child->time;
	}
    }

    /* if no child has terminated, then return 0 */
    if (min_time == MAXTIME)
	return 0;

    prev = pthread;

    /* search the list of children again for the earliest one */
    for (child = pthread->youngest; child; child = child->sibling) {
	if (child->time == min_time
	    && (child->runstate == R_DONE || child->is_zombie)) {
	    /* disown child */
	    if (prev == pthread)
		pthread->youngest = child->sibling;
	    else
		prev->sibling = child->sibling;

	    /* collect cpu time of child */
	    pthread->child_cpu += child->cpu_time + child->child_cpu;

	    /* update parent's elapsed time to that of the child */
	    if (pthread->time < child->time)
		pthread->time = child->time;

	    /*
	     * Set child's parent pointer to NULL so that the child knows
	     * the parent waited on it already, and it can put itself
	     * on the free queue.
	     */
	    child->parent = NULL;

	    /*
	     * If child is not a zombie, then move it to the free pool.
	     * We can't move a zombie child to the free pool, since it
	     * still has to send an E_DONE_THR event, so it is not yet
	     * free to be reallocated. A zombie child will move itself
	     * to the free pool if the parent has waited on it already.
	     */
	    if (child->runstate == R_DONE) {
		/* move child to free pool */
		INLINE_REMOVE (child);
		if (Recycle_threads)
		    INLINE_INSERT_AFTER (&Free_q, child);
		child->runstate = R_FREE;
	    }

	    return child->pid;
	}
	else
	    prev = child;
    }

    /* if we get here, it's an error */
    fatal ("wait_for_child: can't find earliest dead child\n");
    return 0;
}

/*
 * Child thread should wake up its parent:
 *   1) if it satisfies the WAIT_FOR_END condition
 *   2) if it is the last child of the parent
 * If parent is in R_WAIT state after event_check_children, the child
 * must remove itself from the parent list, since the parent expects
 * all children to finish when it wakes up.
 */
int
wakeup_parent (thread_ptr pthread)
{
    thread_ptr parent;
    int to_wakeup_parent;
    thread_ptr prev;

    parent = pthread->parent;
#ifdef DEBUG
    if (parent == NULL)
	fatal ("wakeup_parent:  parent of thread %d should not be null\n",
	       pthread->pid);
#endif /* DEBUG */

    /* check if the parent is blocked on WAIT_FOR_END and should wakeup */
    if (--parent->num_wait_for_end == 0)
	to_wakeup_parent = 1;
    else {
	/* check if parent is waiting for children to finish */
	if (parent->runstate == R_WAIT) {
	    /* remove myself from the parent's list */
	    if (pthread == parent->youngest)
		parent->youngest = pthread->sibling;
	    else {
		for (prev = parent->youngest; prev->sibling != pthread;
		     prev = prev->sibling)
		    ;
		prev->sibling = pthread->sibling;
	    }
	    /* check if I am the parent's last child */
	    if (parent->youngest == NULL)
		to_wakeup_parent = 1;
	    else
		to_wakeup_parent = 0;
	}
	else {
	    /* parent has not reached event_check_children */
	    to_wakeup_parent = 0;
	}
    }
    /* return 1 to wakeup the parent or 0 to not to wake it up */
    return to_wakeup_parent;
}

void
sleep_thr (thread_ptr pthread)
{
    INLINE_REMOVE (pthread);
    INLINE_ENQUEUE (&Sleep_q, pthread);
    pthread->runstate = R_SLEEP;
}

/* Block the thread and put it on the specified wait queue. */
void
block_thr (qnode_ptr q, thread_ptr pthread)
{
    INLINE_REMOVE (pthread);
    INLINE_ENQUEUE (q, (qnode_ptr) pthread);
    pthread->runstate = R_BLOCK;
}

/* dequeue the first waiting thread and make it runnable */
thread_ptr
unblock_thr (thread_ptr q)
{
    thread_ptr pthread;

    INLINE_DEQUEUE (q, pthread);
    if (pthread) {
	INLINE_ENQUEUE (&Run_q, pthread);
	pthread->runstate = R_RUN;
    }
    return pthread;
}

/* This routine simulates the object code instructions up to the next event */
next_event (task_ptr ptask)
{
    thread_ptr pthread;
    int save_pid;
    mint_time_t save_time;
    task_ptr dbg_next, dbg_prev;
    void (*event_func) (task_ptr, thread_ptr);

#ifdef DEBUG
    /* fill the task with a garbage value (-5) to help catch bugs */
    save_pid = ptask->pid;
    save_time = ptask->time;
    dbg_next = ptask->dbg_next;
    dbg_prev = ptask->dbg_prev;
    memset (ptask, -5, sizeof (task_t));
    ptask->pid = save_pid;
    ptask->time = save_time;
    ptask->dbg_next = dbg_next;
    ptask->dbg_prev = dbg_prev;
#endif

    if (ptask->pid < 0 || ptask->pid >= Max_nprocs)
	fatal ("next_event: task pid is bad (%d)\n", ptask->pid);
    pthread = &Threads[ptask->pid];
    pthread->time = ptask->time;

    /* Set the runstate of a thread to R_RUN. */
    WAKEUP (pthread);

    /*
     * The runstate of a thread remains R_RUN if it needs to run
     * immediately after the event_func.  In that case, context switch
     * to it.  On the other hand, the runstate changes to R_SLEEP to
     * break out of this loop in order to schedule a task for
     * execution in sched_loop.
     */
    while (pthread->runstate == R_RUN) {
	/*
	 *
	 * Context switch from mint thread to application thread and
	 * allow it to run local instructions generate an external
	 * event and context switch to Mint.
	 */
	if (pthread->started)
	    (*ctxsw_fn) (psim_thread, pthread, &pcur_thread);
	else {
	    ctxsw_first (psim_thread, pthread, &pcur_thread);
	    pthread->started = 1;
	}
	/*
	 *
	 * At this point, the thread has updated its event structure
	 * with info such as event type, address, data, etc and
	 * context switched to the simulator thread.  Now, the
	 * simulator thread uses the event type to access the event
	 * table to get the associated OP function. 
	 */
	event_func = Event_tab[TABLE_OFFSET (pthread->pevent->type)];

	/*
	 *
	 * Execute OP macro event function to set up fields in event
	 * and thread structures.  Basically, it sets the associated
	 * event record to the new event and assigns to ptask->ufunc
	 * the corresponding backend function.
	 */
	(event_func) (ptask, pthread);
    }

    /* Set the ptask time to the current time of pthread. */
    ptask->time = pthread->time;
    pthread->pevent->time = pthread->time;

    /*
     * Reset the pevent pointer in the task structure in case
     * the back end changed the pid for the task.
     */
    ptask->pevent = pthread->pevent;
    ptask->priority = DEF_PRIORITY;
    ptask->tstack = NULL;
    task_insert (ptask);
    return T_YIELD;
}

/* used with fake events instead of the assembley routine ctxsw */
void
fake_ctxsw (thread_ptr dummy1, thread_ptr pthread, thread_ptr * dummy2)
{
    event_ptr pevent;
    fake_event_ptr pfevent;

    pevent = pthread->pevent;
    if (pthread->index == -1)
	fatal ("no more fake events while thread has not terminated\n");
    /*
     * pthread->index points to the next fake event for pthread in the
     * fevent array */
    pfevent = &fevent[pthread->index];
    pthread->index = pfevent->next;
    if (pthread->time < pfevent->time)
	pthread->time = pfevent->time;
    pevent->type = pfevent->type;
    pevent->paddr = pfevent->addr;
    pevent->arg1 = pfevent->addr;
    pevent->utype = pfevent->utype;
}

/*
 * If memory reference events are being generated, this function is called
 * before a load instruction is simulated.
 */
OP (event_read)
{
    event_ptr pevent;
    long vaddr, paddr;

#ifdef DEBUG_VERBOSE
    debug_print_event (ptask, pthread, pthread->pevent, "event_read");
#endif /* DEBUG_VERBOSE */
    pevent = pthread->pevent;
    vaddr = pevent->vaddr;
    if (IS_SHARED (vaddr)) {
	paddr = SMAP (vaddr);
	pevent->type = pevent->type | E_SHARED;
    }
    else {
	paddr = PMAP (vaddr);
    }
    pevent->paddr = paddr;
    pthread->paddr = paddr;

    ptask->ufunc = sim_read;
    SLEEP (pthread);
}

OP (event_write)
{
    event_ptr pevent;
    long vaddr, paddr;

#ifdef DEBUG_VERBOSE
    debug_print_event (ptask, pthread, pthread->pevent, "event_write");
#endif /* DEBUG_VERBOSE */
    pevent = pthread->pevent;
    vaddr = pevent->vaddr;
    if (IS_SHARED (vaddr)) {
	paddr = SMAP (vaddr);
	pevent->type = pevent->type | E_SHARED;
    }
    else {
	paddr = PMAP (vaddr);
    }
    pevent->paddr = paddr;
    pthread->paddr = paddr;

    ptask->ufunc = sim_write;
    SLEEP (pthread);
}

/*
 * This routine generates an arbitrary event. It is invoked from within
 * the object program (not from the simulator). The macro prototype is:
 *
 * USER(arg1,arg2,arg3,arg4);
 */
OP (event_user)
{
#ifdef DEBUG_VERBOSE
    debug_print_event (ptask, pthread, pthread->pevent, "event_user");
#endif /* DEBUG_VERBOSE */
    ptask->ufunc = sim_user;
    SLEEP (pthread);
}

/* events for shared memory references.   Currently not used */
OP (event_shr_read)
{
    event_ptr pevent;
    long vaddr, paddr;

#ifdef DEBUG_VERBOSE
    debug_print_event (ptask, pthread, pthread->pevent, "event_shr_read");
#endif /* DEBUG_VERBOSE */
    pevent = pthread->pevent;
    vaddr = pevent->vaddr;
    if (!IS_SHARED (vaddr)) {
	pevent->paddr = PMAP (vaddr);
	if (!Fevents)
	    mem_ref (pevent);
	ptask->ufunc = task_yield;
	SLEEP (pthread);
	return;
    }
    /* use version of function for shared references */
    paddr = SMAP (vaddr);
    pevent->paddr = paddr;
    pevent->type = pevent->type | E_SHARED;
    ptask->ufunc = sim_read;
    SLEEP (pthread);
}

/* events for shared memory references.  Currently not used */
OP (event_shr_write)
{
    event_ptr pevent;
    long vaddr, paddr;

#ifdef DEBUG_VERBOSE
    debug_print_event (ptask, pthread, pthread->pevent, "event_shr_write");
#endif /* DEBUG_VERBOSE */
    pevent = pthread->pevent;
    vaddr = pevent->vaddr;
    if (!IS_SHARED (vaddr)) {
	pthread->paddr = PMAP (vaddr);
	if (!Fevents)
	    mem_ref (pevent);
	ptask->ufunc = task_yield;
	SLEEP (pthread);
	return;
    }
    /* use version of function for shared references */
    paddr = SMAP (vaddr);
    pevent->paddr = paddr;
    pevent->type = pevent->type | E_SHARED;
    ptask->ufunc = sim_write;
    SLEEP (pthread);
}

/*
 * This function is called to force the scheduler out of its run-until-block
 * loop. It has the side-effect of updating the time accumulated in the
 * scheduler loop. This is only used internally in MINT. The back end
 * never sees these events.  May be used by the augmenter.
 */
OP (event_yield)
{
#ifdef DEBUG_VERBOSE
    debug_print_event (ptask, pthread, pthread->pevent, "event_yield");
#endif /* DEBUG_VERBOSE */
    ptask->ufunc = task_yield;
    SLEEP (pthread);
}

void q_debug (qnode_ptr q);

allq_debug ()
{
    int i;

    /*
     * for (i = 0; i <= Maxpid; i++)
     * dump_thread(&Threads[i]); */

    printf ("Run_q\n");
    q_debug ((qnode_ptr) & Run_q);

    printf ("Sleep_q\n");
    q_debug ((qnode_ptr) & Sleep_q);

    printf ("Done_q\n");
    q_debug ((qnode_ptr) & Done_q);

    printf ("Free_q\n");
    q_debug ((qnode_ptr) & Free_q);
}

void
q_check (qnode_ptr q)
{
    int i;
    qnode_ptr pnode, prev;

    if (q == NULL)
	return;
    pnode = q;
    if (pnode->next == pnode) {
	if (pnode->prev != pnode) {
	    printf ("bad prev pointer\n");
	    exit (1);
	}
    }
    else {
	for (i = 0; i <= Maxpid; i++) {
	    prev = pnode;
	    pnode = pnode->next;
	    if (pnode->prev != prev) {
		printf ("pnode->prev bad\n");
		exit (1);
	    }
	    if (pnode == q)
		break;
	}
	if (i > Maxpid) {
	    printf ("too many nodes on queue\n");
	    exit (1);
	}
    }
}

allq_check ()
{
    q_check ((qnode_ptr) & Run_q);
    q_check ((qnode_ptr) & Sleep_q);
    q_check ((qnode_ptr) & Done_q);
    q_check ((qnode_ptr) & Free_q);
}
