/* 
 * Please read the COPYRIGHT notice included at the top level
 */

/*
 * Memory allocation routines.
 *
 * Copyright (C) 1993 by Jack E. Veenstra (veenstra@cs.rochester.edu)
 * 
 * This file is part of MINT, a MIPS code interpreter and event generator
 * for parallel programs.
 * 
 * MINT is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 1, or (at your option)
 * any later version.
 * 
 * MINT is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with MINT; see the file COPYING.  If not, write to
 * the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <stdio.h>
#include "icode.h"
#include "globals.h"
#include "mem_management.h"
/*
 * This is a memory allocator that uses an array of buckets for
 * holding free blocks whose sizes are powers of two.  The memory
 * blocks are taken from a region of memory that is of the same fixed
 * size for each thread. The "malloc region" is initially one block
 * and its size is a power of two. This block is subdivided into
 * smaller powers of two as needed to satisfy calls to malloc(). When
 * freed, blocks are not merged immediately since it is likely that
 * the same size block will be requested in the future. (In the
 * current implementation, blocks are never merged but this could be
 * added to the malloc() routine.)
 *
 * A common problem with power-of-two memory allocators is that they
 * are space-inefficient when allocating a block whose size is a power
 * of two since a first extra bytes in the allocated region are needed
 * to store the size of the region (so that a subsequent call to
 * free() can figure out how big the region is). This wastes about
 * half of the memory since a region twice as large as requested must
 * be allocated just so that the size information can be prepended.
 *
 * To eliminate this problem, the size information is stored in a
 * separate data structure (allocated outside of the malloc
 * region). The data structure used to store allocated blocks is a
 * hash table. This is because free() is given the address of a block
 * and since the information for the block is not stored in the block
 * itself, the entry containing the information must be searched for
 * using only the block address as a key. The most efficient way to do
 * this is with a hash table.
 */

static int Max_bucket;

/* size of hash table for storing allocated blocks */
#define ALLOC_HASH_SIZE 64
static avail_ptr Avail_free;

/*
 * Initialize the available memory pool to start at 'base' for 'size'
 * bytes.
 */
void
malloc_init (ulong base, ulong size)
{
    int i, bucket;
    ulong shift;
    ulong newbase;
    avail_ptr availp;

    /* round base up to a double word boundary */
    newbase = (base + 7) & ~7;
    size -= newbase - base;
    base = newbase;

    /* find bucket such that 2^bucket <= size */
    for (shift = size, bucket = -1; shift; ++bucket)
	shift >>= 1;

    Max_bucket = bucket;
    malloc_avail = (avail_ptr) calloc (bucket + 1, sizeof (avail_t));
    if (malloc_avail == NULL)
	fatal ("malloc_init: out of memory.\n");

    /* initialize the head nodes for all the queues */
    for (i = 0; i <= Max_bucket; i++) {
	malloc_avail[i].address = 0;
	malloc_avail[i].size = 0;
	INLINE_INIT_Q (&malloc_avail[i]);
    }

    malloc_alloc = hash_init (ALLOC_HASH_SIZE, A_ENTRY_SIZE);

    NEW_ITEM (Avail_free, sizeof (struct avail), availp, "malloc_init");
    availp->address = base;
    availp->size = size;
    INLINE_ENQUEUE (&malloc_avail[bucket], availp);
}

/* allocate a block of 'size' bytes */
OP (event_malloc)
{
    avail_ptr availp, p2;
    int i, bucket, shift, leftover, bucket2;
    unsigned int size;
    a_entry_ptr pentry;
    event_ptr pevent;

    pevent = pthread->pevent;
    size = pevent->arg1;
#ifdef DEBUG_VERBOSE
    printf ("mint_malloc( %d bytes )\n", size);
    debug_print_event (ptask, pthread, pthread->pevent, "event_malloc");
#endif

    /* round size up to a double word boundary */
    size = (size + 7) & ~7;

    /* find bucket such that 2^bucket <= size */
    for (shift = size, bucket = -1; shift; ++bucket)
	shift >>= 1;

    /* find the first free block large enough */
    for (i = bucket; i <= Max_bucket; ++i) {
	for (availp = malloc_avail[i].next;
             availp->size;
             availp = availp->next)
	    if (availp->size >= size)
		break;
	if (availp->size)
	    break;
    }

    if (i > Max_bucket) {
	fatal ("event_malloc: out of memory (requested size = %d)\n", size);
    }

    /* remove the free block from the pthread->avail[] array */
    INLINE_REMOVE (availp);

    /* split the block if necessary */
    leftover = availp->size - size;

    if (leftover > 0) {
	/* find bucket2 such that 2^bucket2 <= leftover */
	for (shift = leftover, bucket2 = -1; shift; ++bucket2)
	    shift >>= 1;
	NEW_ITEM (Avail_free, sizeof (struct avail), p2, "event_malloc");
	p2->address = availp->address + size;
	p2->size = leftover;
	INLINE_ENQUEUE (&malloc_avail[bucket2], p2);
    }

    pevent->rval = PUNMAP (availp->address);
    pentry = (a_entry_ptr) hash_insert (malloc_alloc, availp->address);
    pentry->size = size;
    pentry->bucket = bucket;
#if DEBUG_VERBOSE
    printf ("event_malloc: returning 0x%x (phys 0x%x)\n",
	    REGNUM (2), availp->address);
#endif
    FREE_ITEM (Avail_free, availp);
    ptask->ufunc = sim_alloc;
    SLEEP (pthread);
}

/* free a previously allocated block */
OP (event_free)
{
    avail_ptr availp;
    int bucket;
    ulong addr;
    a_entry_ptr pentry;

    addr = pthread->pevent->arg1;
#ifdef DEBUG_VERBOSE
    printf ("event_free( 0x%x phys 0x%x )\n", addr, PMAP (addr));
    debug_print_event (ptask, pthread, pthread->pevent, "event_free");
#endif
    addr = PMAP (addr);
    pentry = (a_entry_ptr) hash_free_entry (malloc_alloc, addr);
    if (pentry == NULL) {
	warning ("event_free: cannot find allocated block at address 0x%x\n",
		 addr);
	/* free returns 0 on failure */
	pthread->pevent->rval = 0;
	return;
    }
    bucket = pentry->bucket;
    NEW_ITEM (Avail_free, sizeof (struct avail), availp, "malloc_init");
    availp->address = addr;
    availp->size = pentry->size;

    /*
     * put freed block at the front of the list so it will get reused
     * first
     */
    INLINE_INSERT_AFTER (&malloc_avail[bucket], availp);

    /* free() returns 1 on success */
    pthread->pevent->rval = 1;
}

/* print out the list of blocks on the free list */
void
malloc_stat ()
{
    int i;
    avail_ptr availp;

    for (i = 0; i <= Max_bucket; ++i) {
	printf ("malloc_avail[%d]:\n", i);
	for (availp = malloc_avail[i].next;
             availp->size;
             availp = availp->next)
	    printf ("  %d", availp->size);
	printf ("\n");
    }
}
