/* 
 * Please read the COPYRIGHT notice included at the top level
 */
/*
 * Routines for creating and scheduling tasks.
 *
 * Copyright (C) 1993 by Jack E. Veenstra (veenstra@cs.rochester.edu)
 * 
 * This file is part of MINT, a MIPS code interpreter and event generator
 * for parallel programs.
 * 
 * MINT is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 1, or (at your option)
 * any later version.
 * 
 * MINT is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with MINT; see the file COPYING.  If not, write to
 * the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <stdio.h>
#include "icode.h"
#include "globals.h"
#include "wheel.h"
/*
 * #include "symtab.h" may be useful in printing simulator function
 * names
 */

static qnode_t Block_q;
task_ptr Task_free;
int Num_tasks;			/* to help find memory leaks */
task_t Active_tasks;		/* head node of list of allocated tasks */
/* extern hash_tab_ptr Func_hash_tab; */

/* function prototypes */
void sched_loop ();
int return_advance (task_ptr ptask);
void record_event (task_ptr ptask);
void print_events ();

void task_insert_before (task_ptr ptask, task_ptr before);

void
execute ()
{
    int pid;

#ifdef DEBUG
    INLINE_INIT_Q_NP (&Active_tasks, dbg_next, dbg_prev);
#endif
    wheel_init ();

    /*
     * create a task for the first process to start at time 0
     * need to set its runstate to R_SLEEP to satisfy the consistency
     * check
     */
    if (Fevents) {
	for (pid = 0; pid < Max_nprocs; pid++) {
	    Threads[pid].runstate = R_SLEEP;
	    unblock_proc (pid, (mint_time_t) 0);
	}
    }
    else {
	pid = 0;
	Threads[pid].runstate = R_SLEEP;
	unblock_proc (pid, (mint_time_t) 0);
    }
    /* initialize the head node for the blocked queue */
    INLINE_INIT_Q (&Block_q);
    sched_loop ();
}

void
sched_loop ()
{
    int advance_thread;
    task_ptr ptask, pnew;
    thread_ptr pthread;
    /*    func_name_ptr fptr; */
    char *fname;

    /* get the first task */
    INLINE_TASK_EXTRACT (ptask);

    while (1) {

#ifdef DEBUG_TASK
	printf ("p%d: event: time %.0f, cpu_time %.0f, type = 0x%x\n",
		pevent->pid, *pevent->time, *pevent->cpu_time, pevent->type);
	printf ("     task: time %.0f\n", ptask->time);
#endif

#ifdef DEBUG
	debug_task (ptask);
#endif
#ifdef RECORD_EVENTS
	record_event (ptask);
#endif
	cur_time = ptask->time;
	advance_thread = (*ptask->ufunc) (ptask);
	if (advance_thread == T_ADVANCE) {
	    /*
	     * All the tasks for this event have completed,
	     * so advance the thread to the next event.
	     */
#ifdef DEBUG
	    /*
	     * Check that the back-end did not incorrectly return T_ADVANCE
	     * for this thread. This check is implemented by comparing
	     * "cur_time" with pthread->time. If cur_time, the time
	     * the back-end function was called, is less than pthread->time,
	     * the time the thread has already advanced to, then the
	     * back-end has made a mistake. Note that the mistake may
	     * have been made on the *first* T_ADVANCE instead of this
	     * one. There is no way for the front-end to know.
	     */
	    pthread = &Threads[ptask->pid];
	    if (cur_time < pthread->time) {
		/*
		 * fptr = (func_name_ptr) hash_lookup(Func_hash_tab,
		 * (ulong) ptask->ufunc);
		 * if (fptr)
		 * fname = fptr->fname;
		 * else */
		fname = "(unknown)";
		fprintf (stderr, "sched_loop: at time %.0f task"
                         " \"%s\" returned T_ADVANCE for process %d "
                         "that is already at time %.0f\n",
			 cur_time, fname, ptask->pid, pthread->time);
		exit (1);
	    }

#endif
	    if (cur_time == ptask->time) {
		/*
		 * If the current time is the same, call next_event()
		 * directly, thereby avoiding extra calls to task_insert()
		 * and task_extract().
		 */
		next_event (ptask);
		/* next_event() reschedules the task, so don't free it */
		INLINE_TASK_EXTRACT (ptask);
#ifdef DEBUG
		if (ptask == NULL)
		    fatal ("sched_loop: Wheel underflow\n");
#endif
	    }
	    else {
		/*
		 * The simulator changed the time of ptask so schedule
		 * the next_event() call for that time in the future.
		 */
		ptask->ufunc = next_event;
		INLINE_TASK_INSERT (ptask);
		INLINE_TASK_EXTRACT (ptask);
#ifdef DEBUG
		if (ptask == NULL)
		    fatal ("sched_loop: Wheel underflow\n");
#endif
	    }
	}
	else if (advance_thread == T_CONTINUE) {
	    /*
	     * If there is a task on the task stack, then schedule it.
	     * Otherwise, treat this as a T_ADVANCE, and call next_event().
	     */
#ifdef DEBUG
	    /*
	     * Check that the back-end did not incorrectly return
	     * T_CONTINUE for this thread.  Note that the mistake may have
	     * been made on an earlier T_CONTINUE instead of this one.
	     * There is no way for the front-end to know.
	     */
	    pthread = &Threads[ptask->pid];
	    if (ptask->tstack == NULL && cur_time < pthread->time) {
		/*
		 * fptr = (func_name_ptr) hash_lookup(Func_hash_tab,
		 * (ulong) ptask->ufunc);
		 * if (fptr)
		 * fname = fptr->fname;
		 * else */
		fname = "(unknown)";
		fprintf (stderr, "sched_loop: at time %.0f task "
                         "\"%s\" returned T_CONTINUE for process %d "
                         "that is already at time %.0f\n",
			 cur_time, fname, ptask->pid, pthread->time);
		exit (1);
	    }

#endif
	    if (cur_time == ptask->time) {
		/*
		 * If the current time is the same, call next_event()
		 * directly, thereby avoiding extra calls to task_insert()
		 * and task_extract().
		 */
		/* First, though, check the task stack for pending tasks. */
		pnew = ptask->tstack;
		if (pnew) {
		    /* there are more tasks to execute before the next event */
#ifdef DEBUG
		    Num_tasks--;
		    INLINE_REMOVE_NP (ptask, dbg_next, dbg_prev);
#endif
		    FREE_ITEM (Task_free, ptask);
		    pnew->time = cur_time;

		    /*
		     * Should we just execute the new task, or do we need
		     * to do a reschedule first? Since the task that just
		     * finished is the one that is furthest behind, we can
		     * be sure that it is still the furthest behind since
		     * the time has not changed. The new task, however, may
		     * have a lower priority than another task at this time.
		     * This is an interface question: should the task on
		     * the stack inherit the priority of the "stackee" or
		     * should the stacked task have its own priority?
		     * If stacked tasks have their own priority, it costs
		     * an extra reschedule. This decision should probably
		     * be moved to the back-end by means of an additional
		     * interface function.
		     */
		    INLINE_TASK_INSERT (pnew);
		    INLINE_TASK_EXTRACT (ptask);
		}
		else {
		    /* there are no tasks on the task stack */
		    next_event (ptask);
		    /* next_event() reschedules the task, so don't free it */
		    INLINE_TASK_EXTRACT (ptask);
#ifdef DEBUG
		    if (ptask == NULL)
			fatal ("sched_loop: Wheel underflow\n");
#endif
		}
	    }
	    else {
		/*
		 * The simulator changed the time of ptask so schedule
		 * the next_event() call for that time in the future.
		 */
		/* First, though, check the task stack for pending tasks. */
		pnew = ptask->tstack;
		if (pnew) {
		    /* there are more tasks to execute before the next event */
		    pnew->time = ptask->time;
#ifdef DEBUG
		    Num_tasks--;
		    INLINE_REMOVE_NP (ptask, dbg_next, dbg_prev);
#endif
		    FREE_ITEM (Task_free, ptask);
		    ptask = pnew;
		}
		else {
		    /* there are no tasks on the task stack */
		    ptask->ufunc = next_event;
		}
		INLINE_TASK_INSERT (ptask);
		INLINE_TASK_EXTRACT (ptask);
#ifdef DEBUG
		if (ptask == NULL)
		    fatal ("sched_loop: Wheel underflow\n");
#endif
	    }
	}
	else {
	    /* This was not the task that advances the thread. */

	    /* If the last task terminated the last thread, then we're done. */
	    if (Nprocs == 0)
		break;

	    /*
	     * this can be a VERY expensive check, so don't include it
	     * with the usual DEBUG checks
	     */
#ifdef SLOW_DEBUG
	    /*
	     * Check that there is at least one task still scheduled
	     * on the time wheel somewhere for this thread. It is
	     * not necessarily an error if there is none, but it
	     * probably is, so we should give a warning.
	     */
	    if (Nprocs > 1 && Threads[ptask->pid].runstate == R_SLEEP
		&& !scheduled_task_exists (ptask->pid)
		&& !sim_exists (ptask->pid)) {
		/*
		 * fptr = (func_name_ptr) hash_lookup(Func_hash_tab,
		 * (ulong) ptask->ufunc);
		 * if (fptr)
		 * fname = fptr->fname;
		 * else */
		fname = "(unknown)";
		fprintf (stderr,
			 "sched_loop: at time %.0f task "
                         "\"%s\" returned T_NO_ADVANCE for process %d\n",
			 cur_time, fname, ptask->pid);
		warning ("sched_loop: process %d has no more "
                         "scheduled tasks\n",
			 ptask->pid);
	    }
#endif

	    if (advance_thread == T_NO_ADVANCE) {
		/* Free this task */
#ifdef DEBUG
		Num_tasks--;
		INLINE_REMOVE_NP (ptask, dbg_next, dbg_prev);
#endif
		FREE_ITEM (Task_free, ptask);
	    }
	    else if (advance_thread >= T_MAX) {
		fatal ("Unexpected return value from back-end: %d\n",
		       advance_thread);
	    }

	    /* Get the next task with the minimum time. */
	    INLINE_TASK_EXTRACT (ptask);
#ifdef DEBUG
	    if (ptask == NULL)
		fatal ("sched_loop: Wheel Underflow\n");
#endif
	}
    }
    if (!wheel_empty ())
	warning ("all processes terminated but not all tasks "
                 "have executed.\n");
}

int
return_advance (task_ptr ptask)
{
    return T_ADVANCE;
}

int
task_fork (task_ptr ptask)
{
    event_ptr pevent, child_event;
    task_ptr child_task;

    sim_fork (ptask);

    pevent = ptask->pevent;

    /* get a pointer to the child event */
    child_event = pevent->pevent;

    /* allocate and initialize a task for the child event */
    NEW_ITEM (Task_free, sizeof (task_t), child_task, "task_fork");
#ifdef DEBUG
    /* fill the task with garbage to help catch bugs */
    memset (child_task, -5, sizeof (task_t));
    Num_tasks++;
    INLINE_ENQUEUE_NP (&Active_tasks, child_task, dbg_next, dbg_prev);
#endif
    child_task->time = ptask->time;
    child_task->priority = 0;
    child_task->pid = child_event->pid;
    child_task->pevent = child_event;
    child_task->uptr = NULL;

    /* Change parent's task function pointer so we don't re-execute E_FORK */
    ptask->ufunc = task_yield;

    /* Give the child a yield event since it has no event yet. */
    child_task->ufunc = task_yield;

    /* insert the child task in the task queue */
    INLINE_TASK_INSERT (child_task);

    /* advance the parent */
    return T_ADVANCE;
}

int
task_block (task_ptr ptask)
{
    event_ptr pevent;

    sim_block (ptask);

    /*
     * Save the ptask pointer in the pevent structure
     * so that we can find it quickly when we unblock ptask.
     */
    pevent = ptask->pevent;
    pevent->sptr = (void *) ptask;

    /* Change the event to a yield so we don't re-execute this function. */
    ptask->ufunc = task_yield;

    /* Put this task on the blocked queue. */
    INLINE_ENQUEUE ((task_ptr) & Block_q, ptask);

    /* return T_YIELD so that the execute_loop doesn't free this task */
    return T_YIELD;
}

/*
 * This function unblocks a task and schedules it. The sim_unblock()
 * function is called with a pointer to the unblocked task. The unblocking
 * task continues execution.
 */
int
task_unblock (task_ptr ptask)
{
    event_ptr pevent;
    task_ptr unblocked;

    pevent = ptask->pevent;
    unblocked = (task_ptr) pevent->pevent->sptr;

    /* reset the task's time */
    unblocked->time = ptask->time;

    /* take the unblocked task off the blocked queue */
    INLINE_REMOVE (unblocked);

    /* change the event type in case sim_unblock() needs it */
    /* unblocked->pevent->type = E_UNBLOCK; */
    sim_unblock (unblocked);

    /* schedule the newly unblocked task */
    INLINE_TASK_INSERT (unblocked);

    /* advance the unblocking thread */
    return T_CONTINUE;
}

/*
 * This function is similar to task_unblock() except that a list of
 * tasks are unblocked instead of a single task. The sim_unblock()
 * fuction is called for each unblocked task with a pointer to that
 * task. The unblocking task continues execution.
 */
int
task_unblock_list (task_ptr ptask)
{
    event_ptr pevent, ptmp, next;
    task_ptr unblocked;

    pevent = ptask->pevent;

    /* the events are linked through the next field */
    for (ptmp = pevent->next; ptmp; ptmp = next) {
	next = ptmp->next;

	/* reset the pointer to NULL for the next call to event_unblock_list */
	ptmp->next = NULL;

	unblocked = (task_ptr) ptmp->sptr;

	/* reset the task's time */
	unblocked->time = ptask->time;

	/* take the unblocked task off the blocked queue */
	INLINE_REMOVE (unblocked);

	/* change the event type in case sim_unblock() needs it */
	/* unblocked->pevent->type = E_UNBLOCK; */
	sim_unblock (unblocked);

	/* schedule the newly unblocked task */
	INLINE_TASK_INSERT (unblocked);
    }

    /* reset the pointer to NULL for the next call to event_unblock_list */
    pevent->next = NULL;

    /* advance the unblocking thread */
    return T_ADVANCE;
}

int
task_yield (task_ptr ptask)
{
    return T_ADVANCE;
}

int
task_terminate (task_ptr ptask)
{
    /*
     * call sim_terminate() after all the tasks for this process
     * have been executed.
     */
    stack_task (ptask, sim_terminate);
    return task_cleanup (ptask);
}

int
task_done (task_ptr ptask)
{
    Nprocs--;
    return T_NO_ADVANCE;
}

int
task_synch_time (task_ptr ptask)
{
    return T_ADVANCE;
}

/*
 * Create a new task to execute function "ufunc" at time "time",
 * with the specified priority.
 */
task_ptr
pri_sched_task (task_ptr ptask, PFTASK ufunc, mint_time_t simtime,
                int priority)
{
    task_ptr pnew;

    /* create a temporary new task */
    NEW_ITEM (Task_free, sizeof (task_t), pnew, "pri_sched_task");
#ifdef DEBUG
    /* fill the task with garbage to help catch bugs */
    memset (pnew, -5, sizeof (task_t));
    Num_tasks++;
    INLINE_ENQUEUE_NP (&Active_tasks, pnew, dbg_next, dbg_prev);
#endif
    pnew->pevent = ptask->pevent;
    pnew->pid = ptask->pid;
    pnew->tstack = ptask->tstack;
    pnew->ufunc = ufunc;
    pnew->time = simtime;
    pnew->priority = priority;

    INLINE_TASK_INSERT (pnew);
    return pnew;
}

/*
 * Create a new task to execute function "ufunc" at time "time",
 * with a default priority.
 */
task_ptr
sched_task (task_ptr ptask, PFTASK ufunc, mint_time_t simtime)
{
    task_ptr pnew;

    /* create a temporary new task */
    NEW_ITEM (Task_free, sizeof (task_t), pnew, "sched_task");
#ifdef DEBUG
    /* fill the task with garbage to help catch bugs */
    memset (pnew, -5, sizeof (task_t));
    Num_tasks++;
    INLINE_ENQUEUE_NP (&Active_tasks, pnew, dbg_next, dbg_prev);
#endif
    pnew->pevent = ptask->pevent;
    pnew->pid = ptask->pid;
    pnew->tstack = ptask->tstack;
    pnew->ufunc = ufunc;
    pnew->time = simtime;
    pnew->priority = DEF_PRIORITY;

    INLINE_TASK_INSERT (pnew);
    return pnew;
}

/*
 * Create a new task to execute function "ufunc" before a task "before"
 * with a default priority. The "before" task must already be scheduled.
 * The newly scheduled task inherits the time and priority from the
 * "before" task.
 */
task_ptr
sched_task_before (task_ptr ptask, task_ptr before, PFTASK ufunc)
{
    task_ptr pnew;

    /* create a temporary new task */
    NEW_ITEM (Task_free, sizeof (task_t), pnew, "sched_task");
#ifdef DEBUG
    /* fill the task with garbage to help catch bugs */
    memset (pnew, -5, sizeof (task_t));
    Num_tasks++;
    INLINE_ENQUEUE_NP (&Active_tasks, pnew, dbg_next, dbg_prev);
#endif
    pnew->pevent = ptask->pevent;
    pnew->pid = ptask->pid;
    pnew->tstack = ptask->tstack;
    pnew->ufunc = ufunc;
    pnew->time = before->time;
    pnew->priority = before->priority;

    task_insert_before (pnew, before);
    return pnew;
}

/*
 * Create a new task to execute whenever ptask returns T_CONTINUE.
 * If ptask does not return T_CONTINUE, then it must have scheduled
 * a task (which may, in turn, have scheduled some tasks, etc.) that
 * will eventually return T_CONTINUE. Only one task in a subtree of
 * related tasks may return T_CONTINUE. When that occurs, the task
 * stack is popped, and the next task is scheduled. If the stack is
 * empty, then next_event() is called.
 *
 * The new task is scheduled at the default priority.
 */
task_ptr
stack_task (task_ptr ptask, PFTASK ufunc)
{
    task_ptr pnew;

    /* create a temporary new task */
    NEW_ITEM (Task_free, sizeof (task_t), pnew, "sched_task");
#ifdef DEBUG
    /* fill the task with garbage to help catch bugs */
    memset (pnew, -5, sizeof (task_t));
    Num_tasks++;
    INLINE_ENQUEUE_NP (&Active_tasks, pnew, dbg_next, dbg_prev);
#endif
    pnew->pevent = ptask->pevent;
    pnew->pid = ptask->pid;
    pnew->tstack = ptask->tstack;
    ptask->tstack = pnew;
    pnew->ufunc = ufunc;
    pnew->priority = DEF_PRIORITY;
    return pnew;
}

/*
 * Create a new task to execute whenever this task returns T_CONTINUE.
 * (See the above comment for stack_task().)
 * The new task is scheduled at the specified priority.
 */
task_ptr
pri_stack_task (task_ptr ptask, PFTASK ufunc, int priority)
{
    task_ptr pnew;

    /* create a temporary new task */
    NEW_ITEM (Task_free, sizeof (task_t), pnew, "sched_task");
#ifdef DEBUG
    /* fill the task with garbage to help catch bugs */
    memset (pnew, -5, sizeof (task_t));
    Num_tasks++;
    INLINE_ENQUEUE_NP (&Active_tasks, pnew, dbg_next, dbg_prev);
#endif
    pnew->pevent = ptask->pevent;
    pnew->pid = ptask->pid;
    pnew->tstack = ptask->tstack;
    ptask->tstack = pnew;
    pnew->ufunc = ufunc;
    pnew->priority = priority;
    return pnew;
}

task_ptr
unblock_proc (int pid, mint_time_t simtime)
{
    task_ptr pnew;

    /* check that the unblocked process is in a consistent state */
    if (Threads[pid].runstate != R_SLEEP)
	fatal ("unblock_proc: process %d runstate is %d, not R_SLEEP.\n",
	       pid, Threads[pid].runstate);

    /* create a temporary new task */
    NEW_ITEM (Task_free, sizeof (task_t), pnew, "unblock_proc");
#ifdef DEBUG
    /* fill the task with garbage to help catch bugs */
    memset (pnew, -5, sizeof (task_t));
    Num_tasks++;
    INLINE_ENQUEUE_NP (&Active_tasks, pnew, dbg_next, dbg_prev);
#endif
    pnew->pevent = Threads[pid].pevent;
    pnew->pid = pid;
    pnew->tstack = NULL;
    pnew->ufunc = next_event;
    pnew->time = simtime;
    pnew->priority = DEF_PRIORITY;

    INLINE_TASK_INSERT (pnew);
    return pnew;
}

task_ptr
unblock_proc_now (int pid)
{
    task_ptr pnew;

    /* check that the unblocked process is in a consistent state */
    if (Threads[pid].runstate != R_SLEEP)
	fatal ("unblock_proc: process %d runstate is %d, not R_SLEEP.\n",
	       pid, Threads[pid].runstate);

    /* create a temporary new task */
    NEW_ITEM (Task_free, sizeof (task_t), pnew, "unblock_proc");
#ifdef DEBUG
    /* fill the task with garbage to help catch bugs */
    memset (pnew, -5, sizeof (task_t));
    Num_tasks++;
    INLINE_ENQUEUE_NP (&Active_tasks, pnew, dbg_next, dbg_prev);
#endif
    pnew->pevent = Threads[pid].pevent;
    pnew->pid = pid;
    pnew->tstack = NULL;
    pnew->ufunc = next_event;
    pnew->priority = DEF_PRIORITY;

    task_insert_current (pnew);
    return pnew;
}

task_ptr
schedule_task (task_ptr pnew)
{
    INLINE_TASK_INSERT (pnew);
    return pnew;
}

void
cancel_task (task_ptr ptask)
{
    INLINE_REMOVE (ptask);
#ifdef DEBUG
    Num_tasks--;
    INLINE_REMOVE_NP (ptask, dbg_next, dbg_prev);
#endif
    FREE_ITEM (Task_free, ptask);
}

/*
 * task_destroy() destroys all the scheduled tasks for a process.
 * This is called when a process receives a signal that terminates that
 * process.
 */
void
task_destroy (int pid)
{
    int i;
    task_ptr pnew;

    /* remove all this process's tasks from the Future queue */
    for (pnew = Future_q->next; pnew != Future_q; pnew = pnew->next)
	if (pnew->pid == pid) {
	    INLINE_REMOVE (pnew);
#ifdef DEBUG
	    Num_tasks--;
	    INLINE_REMOVE_NP (pnew, dbg_next, dbg_prev);
#endif
	    FREE_ITEM (Task_free, pnew);
	}

    /* search backwards, starting at the end of the wheel */
    for (i = WHEEL_SIZE - 1; i >= 0; i--) {
	pnew = Wheel[i].prev;
	/* if the queue is empty, continue with the next queue */
	if (pnew == (task_ptr) & Wheel[i])
	    continue;
	while (pnew != (task_ptr) & Wheel[i]) {
	    if (pnew->pid == pid) {
		INLINE_REMOVE (pnew);
#ifdef DEBUG
		Num_tasks--;
		INLINE_REMOVE_NP (pnew, dbg_next, dbg_prev);
#endif
		FREE_ITEM (Task_free, pnew);
	    }
	    pnew = pnew->prev;
	}
    }
}

/*
 * This function creates a new task and initializes the fields.
 * The function pointer field is also copied. This function may be used
 * by simulators that need to create a task without scheduling it.
 */
task_ptr
new_task (task_ptr ptask)
{
    task_ptr pnew;

    NEW_ITEM (Task_free, sizeof (task_t), pnew, "new_task");
#ifdef DEBUG
    /* fill the task with garbage to help catch bugs */
    memset (pnew, -5, sizeof (task_t));
    Num_tasks++;
    INLINE_ENQUEUE_NP (&Active_tasks, pnew, dbg_next, dbg_prev);
#endif
    pnew->pevent = ptask->pevent;
    pnew->pid = ptask->pid;
    pnew->ufunc = ptask->ufunc;
    pnew->priority = DEF_PRIORITY;
    pnew->tstack = ptask->tstack;
    return pnew;
}

/*
 * This function creates a new task but does not initialize any fields.
 * The intended use of this function is for creating tasks that are
 * not associated with any process or event and do not return T_ADVANCE.
 */
task_ptr
newtask ()
{
    task_ptr pnew;

    NEW_ITEM (Task_free, sizeof (task_t), pnew, "new_task");
#ifdef DEBUG
    /* fill the task with garbage to help catch bugs */
    memset (pnew, -5, sizeof (task_t));
    Num_tasks++;
    INLINE_ENQUEUE_NP (&Active_tasks, pnew, dbg_next, dbg_prev);
#endif
    return pnew;
}

void
free_task (task_ptr ptask)
{
#ifdef DEBUG
    Num_tasks--;
    INLINE_REMOVE_NP (ptask, dbg_next, dbg_prev);
#endif
    FREE_ITEM (Task_free, ptask);
}

/* print the first "top" number of allocated tasks */
void
print_active_tasks (int top)
{
    task_ptr pnext;
    /*    func_name_ptr fptr; */
    char *fname;
    int nprinted;

    nprinted = 0;
    pnext = Active_tasks.dbg_next;
    while (pnext != &Active_tasks) {
	if (top > 0 && nprinted >= top)
	    break;
	nprinted++;
	/*
	 * might be useful for printing sim. fn names
	 * fptr = (func_name_ptr) hash_lookup(Func_hash_tab, (ulong)
	 * pnext->ufunc); 
	 if (fptr)
	 fname = fptr->fname;
	 else */
	fname = "(unknown)";
	fprintf (stderr, "\tp%d time %.1f, pri %d, func 0x%x %s,\n",
	     pnext->pid, pnext->time, pnext->priority, pnext->ufunc, fname);
	pnext = pnext->dbg_next;
    }
}

/*
 * Returns true if there is at least one scheduled task for thread
 * pid. */
scheduled_task_exists (int pid)
{
    task_ptr pnext, pstart, dbg_next;

    pnext = Active_tasks.dbg_next;
    while (pnext != &Active_tasks) {
	dbg_next = pnext->dbg_next;
	if (pid == pnext->pid) {
	    /*
	     * Determine if this task is on the time wheel.
	     * Follow the "next" pointer to see if it leads us to
	     e           * a head node in the time wheel. Avoid dereferencing
	     * a NULL pointer, check that the "next" task is linked
	     * back to us through the "prev" pointer, and avoid
	     * infinite loops.
	     */
	    pstart = pnext;
	    while (pnext->next && ((long) pnext->next & 0x3) == 0
		   && pnext->next->prev == pnext
		   && pnext->next != pstart) {
		/* this task is correctly linked with its neighbors */
		pnext = pnext->next;
		if ((long) pnext >= (long) &Wheel[0]
		    && (long) pnext < (long) &Wheel[WHEEL_SIZE])
		    return 1;
		if (pnext == Future_q)
		    return 1;
	    }
	}
	pnext = dbg_next;
    }
    return 0;
}

#ifdef DEBUG

typedef struct eventrec {
    mint_time_t time;
    int pid;
    long paddr;
    long vaddr;
    long iaddr;
    char *fname;
} eventrec_t, *eventrec_ptr;

#define MAX_EVENTS 1000
static eventrec_t Event_history[MAX_EVENTS];
static int Event_index;

void
record_event (task_ptr ptask)
{
    eventrec_ptr prec;
    event_ptr pevent;
    /*    func_name_ptr fptr; */
    char *fname;

    /* don't record the checkpoint task */
    if (ptask->pid < 0)
	return;
    prec = &Event_history[Event_index];
    pevent = ptask->pevent;
    prec->time = ptask->time;
    prec->pid = ptask->pid;
    prec->paddr = pevent->paddr;
    prec->vaddr = pevent->vaddr;
    /*
     * fptr = (func_name_ptr) hash_lookup(Func_hash_tab, (ulong) ptask->ufunc);
     * if (fptr)
     * prec->fname = fptr->fname;
     * else */
    prec->fname = "(unknown)";
    if (++Event_index == MAX_EVENTS)
	Event_index = 0;
}

void
print_events ()
{
    int i;
    eventrec_ptr prec;

    if (Event_history[Event_index].fname) {
	for (i = Event_index; i < MAX_EVENTS; i++) {
	    prec = &Event_history[i];
	    fprintf (stderr, "p%d %7.0f %8x %8x %s\n",
		     prec->pid, prec->time,
		     prec->paddr, prec->iaddr, prec->fname);
	}
    }
    for (i = 0; i < Event_index; i++) {
	prec = &Event_history[i];
	fprintf (stderr, "p%d %7.0f %8x %8x %s\n",
		 prec->pid, prec->time,
		 prec->paddr, prec->iaddr, prec->fname);
    }
}

#endif /* DEBUG */
